

class DictQuery:
    """
    This is a class, that is supposed to help with the processing of multi layered dictionary structures.
    This is a base class for a state object, that processes one such dictionary at a time, the dictionary being set or
    changed by the 'set' method. The class provides the query dict method, with which a nested value of the dict can
    be acquired by using a standard path query like 'key1/key2/key3'.
    It will be attempted to access the nested structure with this path, but in case there is an exception due to a none
    existing path or something else an exception will be caught and the method 'query_exception' will be executed,
    the query will then return whatever this method returns.
    There is the possibility to specify a default value for every query in case it fails and this is also passed to
    the exception method, wo it could be returned from there, for example

    :ivar dict: The currently processed dict of the object. All the queries will be applied to this dict until it is
        changed
    """
    def __init__(self):
        self.dict = None

    def process(self):
        raise NotImplementedError()

    def set(self, query_dict):
        """
        Sets/changes the currently processed dict of the object

        :param query_dict: The new dict to be processed
        :return: void
        """
        self.dict = query_dict

    def query_exception(self, query, query_dict, default):
        """
        This method is being called, when a query to the dict fails.

        :param query: The query string for which the exception occurred
        :param query_dict: The whole dict with which the exception occurred
        :param default: The default value given to the query method
        :return: The return will be the return of the query method in case of failure
        """
        raise NotImplementedError()

    def query_dict(self, dict_query, default):
        """
        Attempts to extract the value to the given query "path" from the nested dictionary structure, but in case it
        fails the method 'query_exception' will be executed.

        :param dict_query: The path-like query string defining the order of keys to apply to the nested dict
        :param default: A possible default value in case the query fails
        :return: The value form the dict
        """
        try:
            keys = dict_query.split('/')
            current_dict = self.dict

            for key in keys:
                current_dict = current_dict[key]

            return current_dict
        except (KeyError, ValueError) as excpetion:
            return self.query_exception(dict_query, self.dict, default)




from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session, scoped_session
from sqlalchemy import create_engine

from sqlalchemy.pool import StaticPool


def session_refresh(database_singleton):

    def generator(func):

        def wrapper(*args, **kwargs):
            self = args[0]
            try:
                return func(*args, **kwargs)
            finally:
                #self.session.close_all()
                #database_singleton._engine.dispose()
                #self.session = database_singleton.get_session()
                pass

        return wrapper

    return generator


def sqlalchemy_add_children(relationship_attribute, object_list):
    """
    EXAMPLE:
    There is a sqlalchemy model "Shepard" and he has a relationship 'sheeps' to a table to 'Sheep' models, than this
    function serves as a helper, because every Sheep has to be added by calling the 'append' method and in case
    one does not want to make a for loop every time that case occurs, this function simply appends all the given models
    in the list to the sqlalchemy relationship attribute
    :param relationship_attribute: A sqlalchemy relationship attribute of a Model, that is a list
    :param object_list: The list of actual Model objects to be appended to the list
    :return: void
    """
    for obj in object_list:
        relationship_attribute.append(obj)


def sqlalchemy_many_list_tuple(session, model_class, param_list, filter_function, object_function):
    """
    Given a list of parameter objects, this function will return a list of corresponding sqlalchemy model objects, by
    querying the database if there already exist some of the specified entries and then creating the remaining objects.
    The list of all those objects created, which are not represented in the database yet is also returned.

    :param session: The session for the database to use
    :param param_list: A list of objects, that each specify a database entry for the given model class
    :param model_class: The class of the model objects to return objects from
    :param filter_function: A function that accepts one item of the param_list as input and will return a sqlalchemy
        filter argument as return, that is used to possibly retrieve the object from the database, that was specified
        by the param item.
        function(param_item): <sqlalchemy filter argument>
    :param object_function: A function that accepts one item of the param_list as input and will return a model object
        for the given model class, that is based on the information the param item describes the object.#
        function(param_item): model_class
    :return: A tuple of two lists. The first being the list of all the model objects, that are based on the items
        given in the parameter list and the second list contains all the model objects for those model, which are not
        represented in the database yet and would still have to be added.
    """
    # This list will contain all items of the parameter list, that could not be retrieved by database query
    remaining_list = []
    # This list will contain all the actual model objects, either retrieved from the database or created
    object_list = []
    for list_item in param_list:
        # For each parameter item attempting to retrieve the corresponding model from the database by querying with
        # the filter function provided
        results = list(session.query(model_class).filter(filter_function(list_item)))
        if len(results) != 0:
            object_list.append(results[0])
        else:
            remaining_list.append(list_item)

    # This list will contain all the model objects, that have been created using the provided class function, but
    # are not already part of the database
    to_add_list = []
    for remaining_item in remaining_list:
        # Creating the model objects using the function provided
        obj = object_function(remaining_item)
        object_list.append(obj)
        to_add_list.append(obj)

    # Returning the list of all the model objects, that correspond to the input list of parameters and the list of
    # all those model objects, that yet have to be added to the database
    return object_list, to_add_list


def mysql_database_singleton(username, password, host, database):
    """
    This function returns a class. A class, that manages the access to a sqlalchemy(mysql) engine that is based on
    the mysql db credentials/parameters passed to this function.
    The class can be used to access the engine and create sessions, by singleton access, which means the actual
    objects for the database access are only created on the first call to the gets session method but from that
    point on session can be retrieved from the static methods of the class, which means there is no need to
    instanciate objects.

    :param username: The username to use for the database access
    :param password: The password to use for the database access
    :param host: The host of the database. Should be localhost
    :param database: The name of the database schema to use
    :return: The MySQLDatabase class
    """
    class MySQLDatabase:

        _engine = None
        _session_maker = None
        _declarative_base = declarative_base()

        @staticmethod
        def get_base():
            return MySQLDatabase._declarative_base

        @staticmethod
        def get_session():
            if MySQLDatabase._session_maker is None:
                MySQLDatabase._create_session_maker()

            return MySQLDatabase._session_maker()  # type: Session

        @staticmethod
        def create_database(base):
            MySQLDatabase._create_engine()
            base.metadata.create_all(MySQLDatabase._engine)

        @staticmethod
        def _create_session_maker():
            MySQLDatabase._create_engine()

            MySQLDatabase._session_maker = sessionmaker(bind=MySQLDatabase._engine)

        @staticmethod
        def _create_engine():
            engine_string = 'mysql+mysqldb://{}:{}@{}/{}'.format(
                username,
                password,
                host,
                database
            )

            MySQLDatabase._engine = create_engine(
                engine_string,
                connect_args={'connect_timeout': 10},
                poolclass=StaticPool
            )

    return MySQLDatabase


def database_controller(database_class):

    class DatabaseControllerBase:

        DATABASE_SINGLETON = database_class  # type: MySQLDatabase

        def __init__(self):
            # Creating the session object
            self.session = self.DATABASE_SINGLETON.get_session()  # type: Session

        def _add(self, model_class, model_instance, contains_function):
            query = self.session.query(contains_function(model_instance))
            if len(list(query)) == 0:
                # Actually adding the model instance to the session, if it is not already n the database
                self.session.add(model_instance)
                self.session.commit()
                # Returning the model instance
                return model_instance
            else:
                # Returning the found instance of the model already existing in the database
                return query.first()

        def update_session(self):
            """
            This method replaces the session
            :return:
            """
            self.session.close()
            self.session = self.DATABASE_SINGLETON.get_session()

        def _select_all(self, model_class):
            try:
                query = self.session.query(model_class).all()
                # The returned object is not a list as much as it is the sql string used to get from the database.
                # Actually getting from the database by converting to string
                return list(query)
            except Exception as e:
                raise e

        def _select(self, model_class, primary_key):
            try:
                query = self.session.query(model_class).get(primary_key).first()
                return query
            except Exception as e:
                raise e

        def _get_all(self, model_class, parameter_list, filter_function, create_function, adding=True):
            # This list will contain all items of the parameter list, that could not be retrieved by database query
            remaining_list = []
            # This list will contain all the actual model objects, either retrieved from the database or created
            object_list = []
            for list_item in parameter_list:
                # For each parameter item attempting to retrieve the corresponding model from the database by
                # querying with the filter function provided
                results = list(self.session.query(model_class).filter(filter_function(list_item)))
                if len(results) != 0:
                    object_list.append(results[0])
                else:
                    remaining_list.append(list_item)

            # This list will contain all the model objects, that have been created using the provided class function,
            # but are not already part of the database
            to_add_list = []
            for remaining_item in remaining_list:
                # Creating the model objects using the function provided
                obj = create_function(remaining_item)
                object_list.append(obj)
                to_add_list.append(obj)

            # In case the adding flag has been set to true, the session will be used to actually already add all the
            # remaining objects to the database
            if adding:
                self.session.add_all(to_add_list)
                self.session.commit()

            # Returning the list of all the model objects, that correspond to the input list of parameters and the list
            # of all those model objects, that yet have to be added to the database
            return object_list, to_add_list

    # Returning the class
    return DatabaseControllerBase



import MySQLdb
import pathlib
import jinja2
import os

from sqlalchemy import create_engine


def _create_database_connector_mysql(username, password, host):
    """
    Creates a new MySQLdb.connect() object, which is used to acquire cursors to the local mysql database server
    instance and these cursors can then be used to query SQL code to the database.
    :param username: The string username to be used for the authentication
    :param password: The string password for the given user
    :param host: The host of the mysql server.Usually the database is installed locally and therefore will be
        'localhost'
    :return: The MySQLdb.connect() object
    """
    connector = MySQLdb.connect(
        host=host,
        user=username,
        passwd=password
    )
    return connector


def create_database_mysql(username, password, host, name):
    """
    Creates a new database schema with the given name on the mysql server at the given host name, using the given
    login credentials to gain access.
    :param username: The string username of the user of the database
    :param password: The password to the given user
    :param host: The host name for the mysql server to connect to
    :param name: The name of the database to be created
    :return: void
    """
    # Getting a new connector object
    connector = _create_database_connector_mysql(username, password, host)
    # Getting a new cursor from that connector onject
    cursor = connector.cursor()
    # Executing a create database clause through that cursor
    sql = 'CREATE DATABASE {}; COMMIT;'.format(name)
    cursor.execute(sql)


def create_sqlalchemy_mysql(username, password, host, database, declarative):
    """
    If given the database name and the desclarative base used for the sqalchemy models this method will setup
    the database (has to exists already) with all the tables according to the sqlalchemy models
    :param module:
    :param database:
    :param declarative:
    :return: void
    """
    # Creating a mysql engine
    engine_string = 'mysql+mysqldb://{}:{}@{}/{}'.format(
        username,
        password,
        host,
        database
    )
    engine = create_engine(engine_string)
    # The sqlalachemy statement to actually create all the tables
    declarative.metadata.create_all(engine)

    return engine


def installation_database_singleton(username, password, host):

    class Db:

        _instance = None

        @staticmethod
        def create_database(name):
            try:
                cursor = Db.get_cursor()
                cursor.execute('CREATE DATABASE {}; COMMIT;'.format(name))
            except:
                print('COULD NOT CREATE DATABASE {}'.format(name))

        @staticmethod
        def get_cursor():
            conn = Db.get_instance()  # type: MySQLdb.Connection
            cursor = conn.cursor()
            return cursor

        @staticmethod
        def get_instance():
            if Db._instance is None:
                Db.new_instance()

            return Db._instance

        @staticmethod
        def new_instance():
            try:
                connector = MySQLdb.connect(
                    host=host,
                    user=username,
                    passwd=password
                )
                Db._instance = connector
                print('DATABASE CONNECTOR SUCCESSFULLY CREATED')
                return True
            except Exception as e:
                print('DURING THE ATTEMPT TO CREATE A DATABASE CONNECTION THE EXCEPTION OCCURRED\n"{}"'.format(str(e)))
                return False

        @staticmethod
        def test():
            return Db.new_instance()

    return Db


class SetupController:

    def run(self):
        raise NotImplementedError()


class FileSetupController(SetupController):

    def __init__(self, path, file_name, template_path, **kwargs):
        self.path = pathlib.Path(path) / file_name
        with open(template_path, mode='r') as file:
            self.template = jinja2.Template(file.read())
        self.content = self.template.render(**kwargs)

    def run(self):
        if not self.exists():
            self.create()

    def create(self):
        with self.path.open(mode='w+') as file:
            file.write(self.content)
            file.flush()

    def exists(self):
        return self.path.exists() and self.path.is_file()


class PathLookupPrompt:

    def __init__(self, ask_current_path=True, path_type='folder', output_function=print, input_function=input):
        self.ask_current_path = ask_current_path
        self.path_type = path_type
        self.print = output_function
        self.input = input_function

    @property
    def path(self):
        return self.get()

    def get(self):
        """
        Returns the path that was input
        :return: string
        """
        # First asking if the current path wants to be used
        if self.ask_current_path:
            flag, path = self._prompt_current_path()
            return path

        # Getting the path from user input
        path_valid = False
        while not path_valid:
            path = self._prompt_path_input()
            path_valid = self._check_path(path)

        return path

    def _check_path(self, path_string):
        """
        Checks if the path exists and if the path has the correct type. Returns true if everything is fine and false in
        case there is a problem.
        :param path_string: The string path to be checked
        :return: boolean
        """
        path = pathlib.Path(path_string)
        if path.exists():
            # Checking if it is the right type
            # Doing things differently depending on which type of path it is supposed to be, specif. by self.path_type
            if self.path_type == 'folder':
                if path.is_dir():
                    return True
                else:
                    self.print('The given path is not a folder!')
            elif self.path_type == 'file':
                if path.is_file():
                    return True
                else:
                    self.print('The given path is not a file')
        else:
            self.print('The given path does not exist')
        # Returning false at the end when true has not been returned before
        return False

    def _prompt_path_input(self):
        """
        Simply prompts the user to input a path string and returns the input
        :return:
        """
        input_string = self.input(
            '\nPlease enter path: '
        )
        return input_string

    def _prompt_current_path(self):
        """
        Presents the user with the string path of the current working directory and asks if this path should be used
        for the prompted path.
        :return: (decision, cwd_path)
        decision - The boolean value of whether the user wants to use the current working directory as the path or not
        cwd_path - The string path of the current working directory
        """
        self.print('\nYour current working directory is:')
        self.print('{}'.format(os.getcwd()))
        input_string = self.input(
            'Would you like to use this path? (y/n)'
        )

        if input_string.upper().strip(' ') in ['Y', 'YES']:
            return True, os.getcwd()
        else:
            return False, os.getcwd()





from setuptools import setup


setup(
    name='jutily',
    version='0.0.0.18',
    description='',
    url='https://github.com/the16thpythonist/ScopusWp',
    author='Jonas Teufel',
    author_email='jonseb1998@gmail.com',
    license='MIT',
    packages=['jutil'],
    include_package_data=True,
    install_requires=[
        'sqlalchemy>=1.2',
        'jinja2>=2.10',
    ],
    python_requires='>=3, <4'
)